#include "distance.h"

#include <string>

#include <unistd.h>
#include <fcntl.h>

#include <thread>

static constexpr const char *ECHO_GPIO_PIN = "35";
static constexpr const char *TRIGGER_GPIO_PIN = "67";
static constexpr char HIGH = '1';
static constexpr char LOW = '0';

// Assumes:
//   - ECHO_GPIO_PIN and TRIGGER_GPIO_PIN have been exported
//   - ECHO_GPIO_PIN is configured as input
//   - TRIGGER_GPIO_PIN is configured as output

Distance::Distance()
{
    trigger_fd = open((std::string("/sys/class/gpio/gpio") + TRIGGER_GPIO_PIN + "/value").c_str(), O_RDWR);
    if (trigger_fd == -1)
    {
        std::perror("failed to open trigger");
        exit(EXIT_FAILURE);
    }

    echo_fd = open((std::string("/sys/class/gpio/gpio") + ECHO_GPIO_PIN + "/value").c_str(), O_RDONLY);
    if (echo_fd == -1)
    {
        std::perror("failed to open echo");
        exit(EXIT_FAILURE);
    }
}

Distance::~Distance()
{
    if (close(trigger_fd) == -1)
    {
        std::perror("failed to close trigger");
        exit(EXIT_FAILURE);
    }

    if (close(echo_fd) == -1)
    {
        std::perror("failed to close echo");
        exit(EXIT_FAILURE);
    }
}

const bool Distance::read_echo() const
{
    char echo_value;
    if (read(echo_fd, &echo_value, 1) == -1)
    {
        std::perror("failed to read echo file");
        exit(EXIT_FAILURE);
    }
    if (lseek(echo_fd, 0, SEEK_SET) == -1)
    {
        std::perror("failed to seek to beginning of echo file");
        exit(EXIT_FAILURE);
    }

    return echo_value == '1';
}

const double Distance::get_reading() const
{
    // SEND TRIGGER
    if (write(trigger_fd, &HIGH, 1) == -1)
    {
        std::perror("failed to write to trigger");
        exit(EXIT_FAILURE);
    }

    std::this_thread::sleep_for(std::chrono::nanoseconds(10));

    if (write(trigger_fd, &LOW, 1) == -1)
    {
        std::perror("failed to write to trigger");
        exit(EXIT_FAILURE);
    }

    auto real_start = std::chrono::high_resolution_clock().now();
    auto start = real_start;
    auto end = real_start;

    constexpr double SPEED_OF_SOUND = 343.0; // in meters/second
    constexpr uint32_t MAX_TIME = 13000000;  // ns

    // READ ECHO
    while (!read_echo())
    {
        start = std::chrono::high_resolution_clock().now();

        if ((start - real_start).count() > MAX_TIME)
        {
            return std::numeric_limits<double>::infinity();
        }
    }

    while (read_echo())
    {
        end = std::chrono::high_resolution_clock().now();

        if ((end - real_start).count() > MAX_TIME)
        {
            return std::numeric_limits<double>::infinity();
        }
    }

    // COMPUTE DISTANCE
    constexpr auto UNIT_SCALE = 10000000; // ns * m/s

    auto diff = end - start;
    return diff.count() * SPEED_OF_SOUND / UNIT_SCALE / 2;
}
